//Playertypes
datablock PlayerData(PlayerFantasy : PlayerStandardArmor)
{
	canJet = false;
	canFantasy = true;

	uiName = "Fantasy Player";
	showEnergyBar = false;
	rechargeRate = 0;
};


//Wall Jump Emitter
datablock ParticleData(FantasyJumpExplosionParticle)
{
	dragCoefficient      = 8;
	gravityCoefficient   = 2.5;
	inheritedVelFactor   = 0.2;
	lifetimeMS           = 300;
	lifetimeVarianceMS   = 100;
	textureName          = "base/data/particles/chunk";
	spinSpeed			= 10.0;
	spinRandomMin		= -50.0;
	spinRandomMax		= 50.0;

	colors[0]     = "0.8 0.8 0.8 0.1";
	colors[1]     = "0.8 0.8 0.8 0";
	sizes[0]      = 0.6;
	sizes[1]      = 0.8;
	times[0]	  = 0;
	times[1]	  = 1;

	useInvAlpha = true;
};
datablock ParticleEmitterData(FantasyJumpExplosionEmitter)
{
   ejectionPeriodMS = 1;
   ejectionVelocity = 6;
   velocityVariance = 1;
   thetaMin         = 89;
   thetaMax         = 90;
   phiVariance      = 360;
   particles = "FantasyJumpExplosionParticle";
};

//High Jump FX
datablock ExplosionData(FantasyWallRunExplosion)
{
   lifeTimeMS = 150;

   emitter[0] = FantasyJumpExplosionEmitter;
   emitter[1] = PushBroomExplosionEmitter;

   particleDensity = 10;
   particleRadius = 0.2;

   faceViewer     = true;
   explosionScale = "1 1 1";
   
   shakeCamera = false;
   camShakeFreq = "10.0 11.0 10.0";
   camShakeAmp = "1 1 1";
   camShakeDuration = 0.5;
   camShakeRadius = 5.0;
};
datablock ProjectileData(FantasyWallRunProjectile)
{
   explosion = FantasyWallRunExplosion;
};


//Fantasy Global Silent Loop
function FantasyGlobalSilentLoop()
{
	cancel($Fantasy::SilentSchedule);
	for(%i=0;%i<ClientGroup.getCount();%i++)
	{
		%client = ClientGroup.getObject(%i);
		if(!isObject(%player = %client.player) || %player.getState() $= "Dead")
			continue;
		%data = %player.getDatablock();
		if(!%data.canFantasy)
			continue;
		FantasyHealthBar(%player);
		FantasyWallRun(%player,%data);
		%isFantasizer = true;
	}

	if(%isFantasizer)
	{
		$Fantasy::SilentSchedule = scheduleNoQuota(1,0,FantasyGlobalSilentLoop);
		$Fantasy::SilentStarted = true;
	}
	else
		$Fantasy::SilentStarted = false;
}


//Health Display
function FantasyHealthBar(%player)
{
	%data = %player.getDatablock();
	%dmg = %player.getDamageLevel();
	%max = %data.maxDamage;
	%ratio = 100/%max;
	%scale = %dmg/%ratio;
	%percent = 100-%scale;
	%gauge1 = FantasyGauge(%percent);
	%health = "<font:impact:36>\c6" @ %gauge1;
	if(!%player.isBricking)
		%message = "<just:right><font:impact:22>" @ "<br><br><br><br>" @ %health SPC "<font:Impact:28>\c6+ ";
	bottomPrint(%player.client,%message,1);
}
//Bar Gauge
function FantasyGauge(%val,%half,%quarter)
{
	%contA = "-----";
	%contB = "----------";
	if(%half)
	{
		%val = %val/2;
		%contB = "";
	}
	if(%quarter)
	{
		%val = %val/4;
		%contA = "";
		%contB = "";
	}
	if(%val < 5)
		%display = "\c7-----" @ %contA @ %contB;
	else if(%val < 10)
		%display = "-\c7----" @ %contA @ %contB;
	else if(%val < 15)
		%display = "--\c7---" @ %contA @ %contB;
	else if(%val < 20)
		%display = "---\c7--" @ %contA @ %contB;
	else if(%val < 25)
		%display = "----\c7-" @ %contA @ %contB;
	else if(%val < 30)
		%display = "-----\c7" @ %contA @ %contB;
	else if(%val < 35)
		%display = "------\c7----" @ %contB;
	else if(%val < 40)
		%display = "-------\c7---" @ %contB;
	else if(%val < 45)
		%display = "--------\c7--" @ %contB;
	else if(%val < 50)
		%display = "---------\c7-" @ %contB;
	else if(%val < 55)
		%display = "----------\c7" @ %contB;
	else if(%val < 60 && !%half)
		%display = "-----------\c7---------";
	else if(%val < 65 && !%half)
		%display = "------------\c7--------";
	else if(%val < 70 && !%half)
		%display = "-------------\c7-------";
	else if(%val < 75 && !%half)
		%display = "--------------\c7------";
	else if(%val < 80 && !%half)
		%display = "---------------\c7-----";
	else if(%val < 85 && !%half)
		%display = "----------------\c7----";
	else if(%val < 90 && !%half)
		%display = "-----------------\c7---";
	else if(%val < 95 && !%half)
		%display = "------------------\c7--";
	else if(%val < 100 && !%half)
		%display = "-------------------\c7-";
	else if(%val == 100 && !%half)
		%display = "--------------------";
	return %display;
}


//Wall Run
function FantasyWallRun(%player,%data)
{
	//Check for Ground
	%pos = %player.getPosition(); //foot position
	%floor = FantasyBoxCheck(%player,1); //high enough for Wall Run
	
	//Hook for Wall Ejecting
	if(isObject(%floor)) //if too close to ground
		%player.FantasyWallStepNum = 0; //reset Wall Step Num
	
	//No-go conditions
	if(isObject(%floor) || %player.isFantasyWallStep) //if too close to ground, or mid-WallRun
		return; //stop here
	
	//Track Upward Velocity
	%vel = %player.getVelocity(); //player velocity
	%upv = getWord(%vel,2); //vertical component of player vel.
	if(%upv == 0) //if vertical vel. is zero
		return; //stop here
	
	//Create sidal vectors
	%ank = vectorAdd(%pos,"0 0" SPC 0.2); //ankle position
	%fvec = %player.getForwardVector(); //forward vector
	%vec[0] = HataDirCheck(%fvec); //cardinalized forward vector
	%vec[1] = vectorScale(%vec[0],-1); //backward vector
	%vec[2] = findRelativeX(%vec[0]); //rightward vector
	%vec[3] = vectorScale(%vec[2],-1); //leftward vector
	
	//Check all 4 sides
	for(%a=0;%a<4;%a++) //Check for all 4 cardinal directions
	{
		%dir = %vec[%a]; //cardinalized vector (N, S, E, W)
		%isWall = isObject(FantasyBeamCheck(%player,%ank,0.7,$TypeMasks::fxBrickObjectType | $TypeMasks::TerrainObjectType,%dir)); //Do raycast, and return bool for if-hit
		if(%isWall) //if Wall Found
			break; //stop looping here
	}
	if(!%isWall) //if Wall not found
		return; //stop here
	
	if(%vec[1] $= %dir) //if Wall is behind you
		return; //stop here

	//Footstep offsets
	%x = %vec[2]; //set right footstep offset
	if(%player.isFantasyWallLeft) //if Wall Running with Left foot (bool)
	{
		%player.isFantasyWallLeft = false; //set to Right foot (bool)
		%x = %vec[3]; //set left footstep offset
	}
	else //otherwise
		%player.isFantasyWallLeft = true; //set to Left foot (bool)
	
	if(%vec[0] !$= %dir) //if not facing wall (%vec[%a] and %dir are both 3-dimensional vectors)
		%x = "0 0 0"; //nullify offset, for single-side running
	
	//Particle Position offsets
	%yMov = vectorScale(%dir,0.6);
	%xMov = vectorScale(%x,0.3);
	%addA = vectorAdd(%yMov,%xMov);
	%addB = vectorAdd(%pos,%addA);
	
	//Do Particle effects
	FantasyFX(%player,FantasyWallRunProjectile,%addB,%dir,0.5);
	
	//Timeout between steps
	schedule(100,0,eval,%player@".isFantasyWallStep = false;"); //..
	%player.isFantasyWallStep = true; //..
	
	//Buff velocity for Wall Running
	if(%upv > 2 && !%player.isFantasyJumping) //if already moving upwards, and not holding Jump
		%player.setVelocity(vectorAdd(%vel,"0 0" SPC 1)); //..
	
	%player.FantasyWallStepNum++; //count succesful function calls
//	talk("Fantasy Wall Step Number = " SPC %player.FantasyWallStepNum);
	
	//Wall Eject if Pressing Jump
	if(%upv >= 0 && %player.isFantasyJumping && %player.FantasyWallStepNum >= 3) //if not moving downwards, and Holding Jump, and 'Wall Step Number' over..
	{
		%yScale = vectorScale(%dir,-5); //eject off the wall
		%yZAdd = vectorAdd(%yScale,"0 0" SPC 5); //eject upwards, and off the wall
		%player.setVelocity(vectorAdd(%vel,%yzAdd)); //Wall Eject, with upward buff
		serverPlay3D(jumpSound,%pos); //play default Jump Sound
	}
}


//Special FX
function FantasyFX(%player,%data,%pos,%vel,%scale)
{
	%p = new Projectile()
	{
		dataBlock = %data;
		initialPosition = %pos;
		initialVelocity = %vel;
		sourceObject = %player;
		client = %player.client;
		scale = %scale SPC %scale SPC %scale;
	};MissionCleanup.add(%p);
	%p.explode();
}

//Hata's Cardinal-ize Directional Vector
function HataDirCheck(%vec)
{
	//Align (North, South, East, West - to match Brick Grid)
	%vec = vectorNormalize(%vec); //normalize the input vector first
	%vecX = getWord(%vec,0) SPC "0 0"; //make new vector from x component only
	%vecY = 0 SPC getWord(%vec,1) SPC 0; //make new vector from y component only

	%vecXLen = vectorLen(%vecX); //measure length of x vector
	%vecYLen = vectorLen(%vecY); //measure length of y vector

	%dir = vectorNormalize(%vecX); //default dir to normalized x vector
	%dir = %vecYLen > %vecXLen ? vectorNormalize(%vecY) : %dir; //if y vector's length is longer - use normalized y vector, otherwise - keep normalized x vector
	return %dir; //return the newly made cardinal directional vector
}

//Box Search
function FantasyBoxCheck(%player,%height)
{
	%height = %height $= "" ? 0.1 : %height; //define %height if un-argumented
	%half = "0 0" SPC -%height/2; //half of arg. height
	%pos = %player.getPosition(); //player foot position
	%start = vectorAdd(%half,%pos); //select midpoint for box search
	%width = 1.25; //select box search width
	%masks = $TypeMasks::fxBrickObjectType | $Typemasks::TerrainObjectType | $Typemasks::StaticShapeObjectType; //select search object types
	initContainerBoxSearch(%start,%width SPC %width SPC %height,%masks); //do box search with args.
	while(%col = containerSearchNext()) //loop indefinitely, while col exists
	{
		if(isObject(%col) && %col != %player) //if col object found, and col is not player
			return %col; //return col object
	}
}

//Beam Check
function FantasyBeamCheck(%player,%pos,%len,%masks,%vec)
{
	%beam = vectorScale(%vec,%len); //vector with magnitude
	%end = vectorAdd(%pos,%beam); //endpoint
	%ray = containerRayCast(%pos,%end,%masks,%player); //do raycast
	%col = firstWord(%ray); //get col object
	return %col; //return col object
}


package FantasyPlayer
{
	//Player Triggers
	function Armor::onTrigger(%data,%player,%slot,%val)
	{
		if(%slot == 2) //if Jumping
			%player.isFantasyJumping = %val; //store Jump bool to player object
		Parent::onTrigger(%data,%player,%slot,%val); //then parent default fcn
	}
	//on Add
	function Armor::onNewDatablock(%data,%player)
	{
		if(!isObject(%client = %player.client)) //if player client DNE
			return; //stop here
		if(%data.getName() $= "PlayerFantasy" && !$Fantasy::SilentStarted) //if in Fantasy playertype, and silent loop not started
			schedule(1,0,FantasyGlobalSilentLoop); //start global silent loop
		Parent::onNewDatablock(%data,%player); //parent default fcn
	}
	//on Death
	function Armor::onDisabled(%data,%player,%enabled)
	{
		if(%data.canFantasy) //if in Fantasy playertype
			FantasyHealthBar(%player); //update status bar on Death
		Parent::onDisabled(%data,%player,%enabled); //parent default fcn
	}
};activatePackage(FantasyPlayer);


//Hata's Relative Vector
function findRelativeX(%vec)
{
	%hVec = getWords(%vec,0,1) SPC 0; //build vector from horizontal component of input vector
	%nhVec = vectorNormalize(%hVec); //normalized horizontal vector
	return vectorCross(%nhVec,"0 0 1"); //return rightward vector
}